#ifndef INC_LIST_H
#define INC_LIST_H

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

/* Predeclarations */

extern void *stat_malloc(size_t);
extern void stat_free(void *);
extern void *safe_realloc(void *,size_t);
extern void *safe_malloc(size_t);
extern char *safe_strdup(char *);
extern char *safe_strdup_with_slack(char *,size_t);

/* Inlines */

static inline void *__alloc_structure(size_t bytes){
	void *p;

	p = safe_malloc(bytes);
	memset(p,0,bytes);
	return p;
}
 
/* Macros */

#define alloc_structure(x) ((x *)__alloc_structure(sizeof(x)))

#define alloc_structure_stretchable(x,stretch_bytes) ((x *)__alloc_structure(sizeof(x) + stretch_bytes))

//#define free_structure(x) stat_free(x)
#define free_structure(x) free(x)

#define SZ(x) x,strlen(x)

#define SWAP(a,b) do { typeof(a) c = a; a = b; b = c; } while (0)

#define ROR4(a,b,c,d) do { typeof(a) z = d; d = c; c = b; b = a; a = z; } while (0)

#define ROL4(a,b,c,d) do { typeof(a) z = a; a = b; b = c; c = d; d = z; } while (0)

/*+-------------------------------------------------------------------------+*/

/* List functions with same interface as in the Linux kernel. */

/* Basic list primitive */
struct list_head {
	struct list_head *next;
	struct list_head *prev;
};

#define LIST_HEAD_INIT(name) { &(name),&(name) }

#define LIST_HEAD(name) struct list_head name = LIST_HEAD_INIT(name)

#define INIT_LIST_HEAD(ptr) do { (ptr)->next = (ptr); (ptr)->prev = (ptr); } while (0)

/*
 * Insert a new item between two other items.
 *
 * You probably want to use list_add below.  Because the actual anchor
 * (if used) is just like another item, no special cases are needed.
 */
static inline void __list_add(struct list_head *new,
                              struct list_head *prev,
                              struct list_head *next){
	next->prev = new;
	new->next = next;
	new->prev = prev;
	prev->next = new;
}

/*
 * Insert a new item after the specified item.
 *
 * If adding after the anchor, this will become the first list on the list.
 */
static inline void list_add(struct list_head *new,struct list_head *head){
	__list_add(new,head,head->next);
}

/*
 * Insert a new item before the specified item.
 *
 * This is called "tail", because it can be considered "appending to the
 * "end of the list" if adding to the anchor.
 */
static inline void list_add_tail(struct list_head *new,struct list_head *head){
	__list_add(new, head->prev, head);
}

/*
 * Splice a list of items into the list at "head".
 */
static inline void list_splice(struct list_head *list, struct list_head *head)
{
	struct list_head *first = list->next;

	if (first != list) {
		struct list_head *last = list->prev;
		struct list_head *at = head->next;

		first->prev = head;
		head->next = first;

		last->next = at;
		at->prev = last;
	}
}

/*
 * Unlink an item from the list with respect to itself by bringing
 * together the pointers from neighboring items.
 */
static inline void __list_del(struct list_head *prev,
                              struct list_head *next){
	next->prev = prev;
	prev->next = next;
}

/*
 * Unlink an item from the list.
 *
 * NOTE: This does NOT change the actual item unless it's the last
 * item on the list (including the anchor)!  To be able to test
 * whether the item is linked to a list or not later, use
 * INIT_LIST_HEAD after list_del().
 */
static inline void list_del(struct list_head *item){
	__list_del(item->prev, item->next);
}

/*
 * Note: This only works for list items, not the head/anchor!
 *
 * This assumes a char is a byte and that a pointer will fit
 * in an unsigned long.
 */
#define list_entry(ptr,type,member) \
	((type *)((char *)(ptr)-(unsigned long)(&((type *)0)->member)))

/*
 * Is the list empty?
 */
static inline int list_empty(struct list_head *head){
	return (head->next == head);
}

/*
 * Transpose (swap) two items in the list, or two lists if
 * list heads are given, or a list and an item in a list if
 * those are given.
 *
 * This is a lot harder than one might first expect! :)
 *
 * There must be an easier, simplified way to do this.
 * It seems like this should be a natural operation.
 */
static inline void list_transpose(struct list_head *one,struct list_head *two){
	if (list_empty(one)){
		if (list_empty(two))
			return;		/* Both empty: do nothing */
		/* "one" is empty, "two" is not */
		one->next = two->next;
		one->prev = two->prev;
		one->next->prev = one;
		one->prev->next = one;
		two->next = two;
		two->prev = two;
		return;
	}

	if (list_empty(two)){
		/* "two" is empty, "one" is not */
		two->next = one->next;
		two->prev = one->prev;
		two->next->prev = two;
		two->prev->next = two;
		one->next = one;
		one->prev = one;
		return;
	}

	if ((one->next != two) &&
	    (one->prev != two)){	/* Not side-by-side */
		SWAP(one->prev->next,two->prev->next);
		SWAP(one->next->prev,two->next->prev);

		SWAP(one->prev,two->prev);
		SWAP(one->next,two->next);
	} else {			/* Side-by-side */
		if (one->next == two){	/* One before two, two after one */
			SWAP(one->prev->next,two->next->prev);
			ROR4(one->prev,two->prev,two->next,one->next);
		} else {		/* One after two, two before one */
			SWAP(one->next->prev,two->prev->next);
			ROL4(one->prev,two->prev,two->next,one->next);
		}
	}

#if 0
Generic unlink/link function, assuming init is next=>self prev=>self:

prevnode;

SWAP(node->prev,prevnode->next->prev);
SWAP(node->next,prevnode->next);

Or:

SWAP(node->prev,node->next->prev);
SWAP(node->next,node->prev->next);

------------
So, to swap of two elements:

Case 1: Nodes not next to each other
     (a)b(c)
     (b)C(d)
     (c)d(e)
     (d)E(f)
     (e)f(g)

trns (a)b(e)
     (d)C(f)
     (e)d(c)
     (b)E(d)
     (c)f(g)

Case 2: Node one next is node two

     (a)b(c)
     (b)C(d)
     (c)D(e)
     (d)e(f)

     (a)b(d)
     (d)C(e)
     (b)D(c)
     (c)e(f)

Caes 3: Node one prev is node two

	(d)e(f)
	(c)D(e)
	(b)C(d)
	(a)b(c)

	(c)e(f)
	(b)D(c)
	(d)C(e)
	(a)b(d)

#endif

}

/*
 * "for each" list item, do { block }.
 */
#define LIST_FOR_EACH(pos,head) \
	for (pos = (head)->next;pos != (head);pos = pos->next)

#endif
